
module MonadicPComb where

data Error a = Ok a | Erro String

instance (Show a) => Show (Error a) where
  showsPrec _ (Erro s) = showString "ERRO: " . showString s
  showsPrec n (Ok x) = showsPrec n x

data Parse a = P (String -> Error (a,String))

parse :: (Parse a) -> String -> Error a
parse p input 
 = case (aplica p input) of
   Ok (x,_) -> Ok x
   Erro x  -> Erro x
aplica :: (Parse a) -> String -> Error (a,String)
aplica (P p) = p

instance Monad Parse where
 return x = P (\ input -> Ok (x,input))
 p >>= f
  = P $ \input -> 
     case (aplica p input) of
      Ok (a,remaining) -> aplica (f a) remaining
      Erro x -> Erro x

unParse :: String -> Parse ()
unParse s = P  ( \ alpha -> Ok ( () ,s++alpha) )

parseConst :: String -> Parse ()
-- Consome uma constante (String) do input
parseConst c = 
 P (\input -> 
  case (lex input) of 
   ((x,y):_) -> if (x == c) then Ok ((),y) else Erro input
   _         -> Erro input)

parseString :: Parse String 
-- o equivalente ao lex
parseString = 
 P (\input -> 
  case (lex input) of
   ((x,y):_) -> Ok (x,y)
   _         -> Erro input)

parseStringTo :: Char -> Parse String 
-- consome todo o input ate aparecer um determinado caracter
-- da erro se tal nao acontecer
parseStringTo c = 
 P (\input -> 
  case (span (/= c) input ) of
   (x,[]) -> Erro ("Fim detectado antes de " ++ [c])
   (a,b) -> Ok (a,b))

parseStringCond :: (Char -> Bool) -> Parse String
parseStringCond p = P (Ok . (span p))
parseChar :: Parse Char
parseChar = 
 P (\input -> 
  case input of 
   "" -> Erro "Fim de input"
   (i:is) -> Ok (i,is))

parseInt,parseUInt :: (Integral a) => Parse a
parseInt = 
 P (\input -> 
  let (sinais,i') = span (== '-') (dropWhile isSpace input)
      (n,remaining) = span isDigit i'
  in case n of 
   "" -> Erro (input ++ " nao e um inteiro")
   _  -> Ok (numero sinais n,remaining))

parseUInt = 
 P (\input -> 
  let (n,remaining) = span isDigit (dropWhile isSpace input)
  in case n of 
   "" -> Erro (input ++ " nao e um inteiro")
   _  -> Ok (numero "" n,remaining))

parseFrac :: (Fractional a) => Parse a
parseFrac = 
 do { 
   int <- parseInt ;
   ifNot 
     (do { 
      parseConst "." ;
      frac <- parseStringCond (isDigit) ;
      return ((fromInt int) + (fromInt (signum int)) * (mkFrac frac)) })
     (return (fromInt int)) }

----------------- Funcoes auxiliares locais
d2n c = fromInt ((ord c) - ord '0')
numero :: (Integral a) => String -> String -> a
numero ss ns = foldl (const . negate) (foldl (\ n d -> n*10 + (d2n d)) 0 ns) ss
mkFrac :: (Fractional a) => String -> a
mkFrac = foldr (\ x y -> ((d2n x) + y) / 10) 0 

-----------------  Combinadores ----------------
parsePar :: (Parse a) -> (Parse b) -> Parse (a,b)
-- o nome e o tipo diz tudo
parsePar p1 p2 = 
 do { 
  x <- p1 ;
  y <- p2 ;
  return (x,y) }

ifNot :: (Parse a) -> (Parse a) -> Parse a
-- tenta aplicar o primeiro; em caso de erro da o resultado do segundo
ifNot (P p1) (P p2) = 
 P (\input -> 
  case (p1 input) of
   Ok x -> Ok x
   _    -> p2 input)

try :: [Parse a] -> Parse a
try [] = P (\input -> Erro ("nao consigo fazer nada com " ++ input))
try (p:ps) = ifNot p (try ps)

single :: (Parse a) -> Parse [a]
-- so para converter elementos em listas singulares
-- e usado por exemplo quando temos uma lista com separadores
single p = 
 do { 
  x <- p ;
  return [x] }

parseSeq :: (Parse a) -> Parse [a]
-- dado um parser de um tipo, aplica-o enquanto puder 
parseSeq p =
 ifNot 
  (do { 
    a <- p ;
    as <- parseSeq p ;
    return (a:as) }) 
  (return [])

parseSeqSep :: String -> (Parse a) -> Parse [a]
-- igual ao anterior, mas para elementos separados por um separador
parseSeqSep s p =
 ifNot 
  (do { 
   a <- p ;
   parseConst s ;
   as <- parseSeqSep s p ;
   return (a:as) }) 
  (ifNot (single p) (return []))

parseParen :: String -> String -> (Parse a) -> Parse a
-- trata do abrir (ch1) e fechar (ch2) de parentesis
parseParen ch1 ch2 p = 
 do { 
  parseConst ch1 ;
  x <- p ;
  parseConst ch2 ;
  return x }

parseStrP :: Parse String
parseStrP = 
 P (psch . (dropWhile isSpace))
  where 
   psch ('{':input) = 
    let (x,y) = getCharsP 0 input
    in case y of 
     ('}':s) -> Ok (x,s)
     _ -> Erro "Chaveta nao fechada"
   psch input = Erro "Falta {"
   getCharsP _ [] = ([],[])
   getCharsP n ('{':t) = add2fst '{' (getCharsP (n+1) t)
   getCharsP 0 ('}':t) = ([],'}':t)
   getCharsP (n+1) ('}':t) = add2fst '}' (getCharsP n t) 
   getCharsP n (h:t) = add2fst h (getCharsP n t)
   add2fst c (a,b) = (c:a,b)

parsePCurvos = parseParen "(" ")"

parsePRectos = parseParen "[" "]"

parseChaveta = parseParen "{" "}"

parseEither :: (Parse a) -> (Parse b) -> Parse (Either a b)
parseEither p1 p2 = ifNot (p1 >>= return . Left) (p2 >>= return . Right)

parseMaybe :: (Parse a) -> Parse (Maybe a)
parseMaybe p = ifNot (p >>= return . Just) (return Nothing)

showLine s = showString (s ++ "\n")


